""" Message classes for OpenAI Chat API """
from dataclasses import dataclass, field


class Role:
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"


DEFAULT_MESSAGE_ROLE = Role.USER


@dataclass
class Msg:
    role: str = field(default=DEFAULT_MESSAGE_ROLE)
    content: str = field(default="")

    def __str__(self):
        return str(self.content)


@dataclass
class SysMsg(Msg):
    role: str = field(default=Role.SYSTEM, init=False)


@dataclass
class UserMsg(Msg):
    role: str = field(default=Role.USER, init=False)


@dataclass
class AssistantMsg(Msg):
    role: str = field(default=Role.ASSISTANT, init=False)



import asyncio
import openai

from ..configuration import Config, ApiType
from .._prepare_llm_args import prepare_chat_messages, prepare_prompt
from ..types import LLMAsyncFunctionType, LLMFunctionType
from ..wrappers.llm_response_wrapper import LLMResponse
from ..utils import is_chat_model

OPENAI_V1_API = True


def _get_chunk_text(chunk, mode_chat_model: bool):
    # Azure API gives first chunk with empty choices
    if len(chunk.choices) == 0:
        return ""
    choice = chunk.choices[0]
    if mode_chat_model:
        if hasattr(choice, "delta"):
            return getattr(choice.delta, "content", "")
        return ""
    return getattr(choice, "text", "")


async def _a_process_streamed_response(
    response, callbacks: list[callable], chat_model_used: bool
):
    response_text: str = ""
    async for chunk in response:
        if text_chunk := _get_chunk_text(chunk, chat_model_used):
            response_text += text_chunk
            for cb in callbacks:
                if asyncio.iscoroutinefunction(cb):
                    await cb(text_chunk)
                else:
                    cb(text_chunk)
    return LLMResponse(response_text, {})


def _process_streamed_response(
    response, callbacks: list[callable], chat_model_used: bool
):
    response_text: str = ""
    for chunk in response:
        if text_chunk := _get_chunk_text(chunk, chat_model_used):
            response_text += text_chunk
            [cb(text_chunk) for cb in callbacks]
    return LLMResponse(response_text, {})


def _prepare_llm_arguments(config: Config, kwargs: dict):
    args = {**config.LLM_DEFAULT_ARGS, **kwargs}
    args["model"] = args.get(
        "model",
        args.get("deployment_id", config.LLM_DEPLOYMENT_ID or config.MODEL)
        if config.LLM_API_TYPE == ApiType.AZURE
        else config.MODEL,
    )
    callbacks: list[callable] = args.pop("callbacks", [])
    if "callback" in args:
        cb = args.pop("callback")
        if cb:
            callbacks.append(cb)
    args["stream"] = bool(callbacks)
    return args, {"callbacks": callbacks}


def make_llm_functions(config: Config) -> tuple[LLMFunctionType, LLMAsyncFunctionType]:
    # _configure_open_ai_package(config)

    if config.LLM_API_TYPE == ApiType.AZURE:
        connection_type = openai.AzureOpenAI
        async_connection_type = openai.AsyncAzureOpenAI
        params = dict(
            api_key=config.LLM_API_KEY,
            azure_endpoint=config.LLM_API_BASE,
            api_version=config.LLM_API_VERSION,
        )
    else:
        connection_type = openai.OpenAI
        async_connection_type = openai.AsyncOpenAI
        params = dict(
            api_key=config.LLM_API_KEY,
            base_url=config.LLM_API_BASE,
        )

    _connection = connection_type(**params)
    _async_connection = async_connection_type(**params)

    async def allm(prompt, **kwargs):
        args, options = _prepare_llm_arguments(config, kwargs)
        if is_chat_model(args["model"]):
            # _connection.chat.completions.create()

            response = await _async_connection.chat.completions.create(
                messages=prepare_chat_messages(prompt), **args
            )
            if args["stream"]:
                return await _a_process_streamed_response(
                    response, options["callbacks"], chat_model_used=True
                )

            for cb in options["callbacks"]:
                cb(response.choices[0].message.content)
            return LLMResponse(response.choices[0].message.content, response.__dict__)

        response = await _async_connection.completions.create(
            prompt=prepare_prompt(prompt), **args
        )
        if args["stream"]:
            return await _a_process_streamed_response(
                response, options["callbacks"], chat_model_used=False
            )

        return LLMResponse(response.choices[0].text, response.__dict__)

    def llm(prompt, **kwargs):
        args, options = _prepare_llm_arguments(config, kwargs)
        if is_chat_model(args["model"]):
            response = _connection.chat.completions.create(
                messages=prepare_chat_messages(prompt), **args
            )
            if args["stream"]:
                return _process_streamed_response(
                    response, options["callbacks"], chat_model_used=True
                )

            for cb in options["callbacks"]:
                cb(response.choices[0].message.content)
            return LLMResponse(response.choices[0].message.content, response.__dict__)

        # Else (if it is text completion model)
        response = _connection.completions.create(prompt=prepare_prompt(prompt), **args)
        if args["stream"]:
            return _process_streamed_response(
                response, options["callbacks"], chat_model_used=False
            )

        return LLMResponse(response.choices[0].text, response.__dict__)

    return llm, allm



import os
import jinja2
from ..types import TplFunctionType


def make_jinja2_env(env) -> jinja2.Environment:
    return jinja2.Environment(
        autoescape=env.config.JINJA2_AUTO_ESCAPE,
        loader=jinja2.ChoiceLoader(
            [jinja2.FileSystemLoader(env.config.PROMPT_TEMPLATES_PATH)]
        ),
    )


def make_tpl_function(env) -> TplFunctionType:
    def tpl(file: os.PathLike[str] | str, **kwargs) -> str:
        return env.jinja_env.get_template(file).render(**kwargs)

    return tpl



from colorama import Fore, init

init(autoreset=True)
_input = input


def info(*args, color=Fore.LIGHTYELLOW_EX, **kwargs):
    print(*[color + i for i in args], **kwargs)


def debug(msg):
    info(msg, color=Fore.BLUE)


def error(*args, **kwargs):
    print(*[Fore.RED + i for i in args], **kwargs)


def ask_yn(msg):
    return "y" in input(msg + " (y/n)").lower().strip()


def ask_choose(msg, variants: list):
    i = 0
    if isinstance(variants, list):
        for item in variants:
            i += 1
            print(f"\t{Fore.MAGENTA}{i}:{Fore.RESET}\t{item}")
    while True:
        i = input(f"{msg} {Fore.MAGENTA}[1-{len(variants)}]{Fore.RESET}: ")
        if not i.isdigit():
            error("Not a number")
            continue
        i = int(i) - 1
        if i >= len(variants) or i < 0:
            error("Incorrect number")
            continue
        break

    item = variants[int(i)]
    return item



from abc import ABC, abstractmethod
from dataclasses import dataclass

from ..utils import ExtendedString


class SearchResult(ExtendedString):
    """
    String containing the search result with additional information in attributes

    Attributes:
        id (str): document (text) identifier in embedding database
        distance (float): The distance between the query and the search result
        metadata (dict): A dictionary containing document metadata
    """

    id: str
    distance: float
    metadata: dict


@dataclass
class AbstractEmbeddingDB(ABC):
    """
    Base class for embedding databases
    """

    @abstractmethod
    def search(
        self,
        collection: str,
        query: str | list,
        n_results: int = 5,
        where: dict = None,
        **kwargs,
    ) -> list[str | SearchResult]:
        """
        Similarity search

        Args:
            collection (str): collection name
            query (str | list): query string or list of query strings
            n_results (int): number of results to return
            where (dict): filter results by metadata
            **kwargs: additional arguments
        """

    @abstractmethod
    def get_all(self, collection: str) -> list[str | SearchResult]:
        """Return all documents in the collection"""

    def save(self, collection: str, text: str, metadata: dict = None):
        """Save a single document in the collection"""
        self.save_many(collection, [(text, metadata)])

    @abstractmethod
    def save_many(self, collection: str, items: list[tuple[str, dict] | str]):
        """Save multiple documents in the collection"""

    @abstractmethod
    def clear(self, collection: str):
        """Clear the collection"""

    def find_one(self, collection: str, query: str | list) -> str | SearchResult | None:
        """
        Find most similar document in the collection

        Returns:
            Most similar document or None if collection is empty
        """
        return next(iter(self.search(collection, query, 1)), None)



import importlib.metadata

_openai_version = importlib.metadata.version("openai")

if _openai_version.startswith("1."):
    from ._openai_llm_v1 import make_llm_functions
else:
    from ._openai_llm_v0 import make_llm_functions

__all__ = ["make_llm_functions"]



from .._llm_functions import allm, llm
from ..utils import ExtendedString


class PromptWrapper(ExtendedString):
    def to_llm(self, **kwargs):
        """
        Send prompt to Large Language Model, see `llm`
        """
        return llm(self, **kwargs)

    async def to_allm(self, **kwargs):
        """
        Send prompt to Large Language Model asynchronously, see `allm`
        """
        return await allm(self, **kwargs)



"""
File storage functions
"""
import fnmatch
import json
import os
import shutil
from pathlib import Path
import chardet

from ._env import config
from .utils import file_link, list_files


class Storage:
    @property
    def path(self) -> Path:
        return Path(config().STORAGE_PATH)

    @property
    def default_ext(self) -> str | None:
        ext = config().STORAGE_DEFAULT_FILE_EXT
        if ext and not ext.startswith("."):
            ext = "." + ext
        return ext

    def file_link(self, file_name: str | Path) -> str:
        """Returns file name in format displayed in PyCharm console as a link."""
        return file_link(self.path / file_name)

    @property
    def default_encoding(self) -> str:
        return config().DEFAULT_ENCODING

    def exists(self, name: str | Path) -> bool:
        return (self.path / name).exists()

    def abs_path(self, name: str | Path) -> Path:
        if os.path.isabs(name):
            return Path(name)
        return self.path / name

    def read(self, name: str | Path, encoding: str = None):
        name = str(name)
        encoding = encoding or self.default_encoding
        if not os.path.isabs(name) and not name.startswith("./"):
            if "." in name:
                parts = name.split(".")
                name = ".".join(parts[:-1])
                ext = "." + parts[-1]
            else:
                ext = self.default_ext
                if not self.exists(f"{name}{ext}"):
                    ext = ""
            name = f"{self.path}/{name}{ext}"
        if encoding is None:
            with open(name, "rb") as f:
                rawdata = f.read()
            result = chardet.detect(rawdata)
            encoding = result["encoding"]
            return rawdata.decode(encoding)

        with open(name, "r", encoding=encoding) as f:
            return f.read()

    def write_json(
        self,
        name: str | Path,
        data,
        rewrite_existing: bool = True,
        backup_existing: bool = True,
        ensure_ascii: bool = False,
    ):
        serialized_data = json.dumps(data, indent=4, ensure_ascii=ensure_ascii)
        return self.write(name, serialized_data, rewrite_existing, backup_existing)

    def read_json(self, name: str | Path, default=None):
        try:
            return json.loads(self.read(name))
        except FileNotFoundError as e:
            if default is not None:
                return default
            raise e

    def delete(self, target: str | Path | list[str | Path]):
        """
        Removes the file or directory specified by `path` within the `storage_path` if exists.
        """
        if isinstance(target, list):
            for t in target:
                self.delete(t)
            return
        path = (self.path / target).resolve()
        if not path.exists():
            return
        if path.is_dir():
            shutil.rmtree(path)
        else:
            os.remove(path)

    def write(
        self,
        name: str | Path,
        content: str = None,
        rewrite_existing: bool = None,
        backup_existing: bool = None,
        encoding: str = None,
        append: bool = False,
    ) -> str | os.PathLike:
        """
        :return: str File name for further usage
        """
        if rewrite_existing is None:
            rewrite_existing = True
        if backup_existing is None:
            backup_existing = not append
        encoding = encoding or self.default_encoding
        if content is None:
            content = name
            name = f"out{self.default_ext}"

        base_name = Path(name).with_suffix("")
        ext = Path(name).suffix or self.default_ext

        file_name = f"{base_name}{ext}"
        if (self.path / file_name).is_file() and (
            backup_existing or not rewrite_existing
        ):
            counter = 1
            while True:
                file_name1 = f"{base_name}_{counter}{ext}"  # noqa
                if not (self.path / file_name1).is_file():
                    break
                counter += 1
            if not rewrite_existing:
                file_name = file_name1
            elif backup_existing:
                os.rename(self.path / file_name, self.path / file_name1)
        (self.path / file_name).parent.mkdir(parents=True, exist_ok=True)
        if append:
            with (self.path / file_name).open(mode="a", encoding=encoding) as file:
                file.write(content)
        else:
            (self.path / file_name).write_text(content, encoding=encoding)
        return file_name

    def clean(self, path: str | Path):
        """
        Removes the directory specified by `path` within the `storage_path`.
        :raises ValueError: If the path is outside the storage area.
        @deprecated use storage.delete() instead
        """
        full_path = (self.path / path).resolve()

        # Verify that the path is inside the storage_path
        if self.path.resolve() not in full_path.parents:
            raise ValueError("Cannot delete directories outside the storage path.")

        if full_path.exists() and full_path.is_dir():
            shutil.rmtree(full_path)

    def list_files(
        self,
        target_dir: str | Path = "",
        exclude: list[str | Path] = None,
        relative_to: str | Path = None,
        absolute: bool = False,
        posix: bool = False,
    ) -> list[Path | str]:
        """
        Lists files in a specified directory, excluding those that match given patterns.

        Args:
            target_dir (str | Path): The directory to search in.
            exclude (list[str | Path], optional): Patterns of files to exclude.
            relative_to (str | Path, optional): Base directory for relative paths.
                If None, paths are relative to `target_dir`. Defaults to None.
            absolute (bool, optional): If True, returns absolute paths. Defaults to False.
            posix (bool, optional): If True, returns posix paths. Defaults to False.
        """
        target_dir = self.path / target_dir
        return list_files(
            target_dir=target_dir,
            exclude=exclude,
            relative_to=relative_to,
            absolute=absolute,
            posix=posix,
        )

    def copy(self, src: str | Path, dest: str | Path, exclude=None):
        """
        Copy a file or folder from src to dest, overwriting content,
        but skipping paths in exceptions.
        Supports Unix shell-style wildcards in exceptions. Accepts Path objects.

        Args:
            src (Path): Source file or directory Path object.
            dest (Path): Destination file or directory Path object.
            exclude (list of str, optional):
                List of Unix shell-style wildcard patterns relative to src.
                These paths will be excluded from the copy. Defaults to None.
        """
        src = self.path / Path(src)
        dest = self.path / Path(dest)
        exclude = exclude or []
        if src.is_dir():
            dest.mkdir(parents=True, exist_ok=True)
            files = list_files(src, exclude)
            for f in files:
                if (src / f).is_dir():
                    (dest / f).mkdir(parents=True, exist_ok=True)
                elif (src / f).is_file():
                    (dest / f).parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy2(src / f, dest / f)
                else:
                    raise ValueError(f"{src / f} is not a file or directory")
        elif src.is_file():
            if not any(fnmatch.fnmatch(src.name, pattern) for pattern in exclude):
                if dest.is_dir():
                    dest = dest / src.name
                dest.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy2(src, dest)
        else:
            raise ValueError(f"{src} is not a file or directory")


storage = Storage()
"""
File system operations within the storage folder.

See `Storage` for details.

Related configuration options:

    - `microcore.config.Config.STORAGE_PATH`
    - `microcore.config.Config.DEFAULT_ENCODING`
"""



"""
ai_module: ai_func
descr: Allows to describe python functions for LLM
"""
import ast
import inspect
from typing import Dict, Any
import docstring_parser
from .. import tpl


def func_arg_comments(func):
    func_source = inspect.getsource(func)
    module = ast.parse(func_source)
    func_def = module.body[0]

    # Get comments in the function's arguments
    arg_comments = {}
    for arg in func_def.args.args:
        # Get the line of code where this argument is defined
        arg_line = func_source.split("\n")[arg.lineno - 1]
        # If there's a comment in this line, get it
        if "#" in arg_line:
            comment = arg_line.split("#")[1].strip()
            arg_comments[arg.arg] = comment
        else:
            arg_comments[arg.arg] = ""

    return arg_comments


def func_metadata(func) -> Dict[str, Any]:
    metadata = dict(
        name=func.__name__,
        description=inspect.getdoc(func),
        args={},
    )

    for name, param in inspect.signature(func).parameters.items():
        # Store parameter info
        metadata["args"][name] = {"kind": str(param.kind)}
        if param.default != param.empty:
            if isinstance(param.default, str):
                metadata["args"][name]["default"] = '"' + param.default + '"'
            else:
                metadata["args"][name]["default"] = param.default
        else:
            metadata["args"][name]["default"] = "NOT_SET"
        if param.annotation != param.empty:
            param_type = (
                str(param.annotation)
                .replace("typing.", "")
                .replace("<class '", "")
                .replace("'>", "")
            )
            metadata["args"][name]["type"] = param_type

    arg_comments = func_arg_comments(func)
    for name, val in metadata["args"].items():
        val["comment"] = arg_comments[name]

        # Parse docstring
    parsed_docstring = docstring_parser.parse(inspect.getdoc(func))

    # Add descriptions from parsed docstring to parameters
    for param in parsed_docstring.params:
        if param.arg_name in metadata["args"]:
            metadata["args"][param.arg_name]["docstr"] = param.description

    return metadata


def describe_ai_func(func):
    metadata = func_metadata(func)
    return tpl("python_ai_func.j2", **metadata)



import asyncio
import openai
import openai.util

from ..configuration import Config, ApiType
from .._prepare_llm_args import prepare_chat_messages, prepare_prompt
from ..types import LLMAsyncFunctionType, LLMFunctionType
from ..wrappers.llm_response_wrapper import LLMResponse
from ..utils import is_chat_model


def _get_chunk_text(chunk, mode_chat_model: bool):
    # Azure API gives first chunk with empty choices
    choice = chunk.choices[0] if len(chunk.choices) else {}

    if mode_chat_model:
        return choice.get("delta", {}).get("content", "")

    return choice.get("text", "")


async def _a_process_streamed_response(
    response, callbacks: list[callable], chat_model_used: bool
):
    response_text: str = ""
    async for chunk in response:
        if text_chunk := _get_chunk_text(chunk, chat_model_used):
            response_text += text_chunk
            for cb in callbacks:
                if asyncio.iscoroutinefunction(cb):
                    await cb(text_chunk)
                else:
                    cb(text_chunk)
    return LLMResponse(response_text, {})


def _process_streamed_response(
    response, callbacks: list[callable], chat_model_used: bool
):
    response_text: str = ""
    for chunk in response:
        if text_chunk := _get_chunk_text(chunk, chat_model_used):
            response_text += text_chunk
            [cb(text_chunk) for cb in callbacks]
    return LLMResponse(response_text, {})


def _configure_open_ai_package(config: Config):
    try:
        api_type = config.LLM_API_TYPE
        openai.util.ApiType.from_str(api_type)
    except openai.error.InvalidAPIType:
        api_type = ApiType.OPEN_AI

    openai.api_type = api_type
    openai.api_key = config.LLM_API_KEY
    openai.api_base = config.LLM_API_BASE
    openai.api_version = config.LLM_API_VERSION


def _prepare_llm_arguments(config: Config, kwargs: dict):
    args = {**config.LLM_DEFAULT_ARGS, **kwargs}

    args["model"] = args.get("model", config.MODEL)

    if config.LLM_API_TYPE == ApiType.AZURE:
        args["deployment_id"] = args.get("deployment_id", config.LLM_DEPLOYMENT_ID)

    callbacks: list[callable] = args.pop("callbacks", [])
    if "callback" in args:
        cb = args.pop("callback")
        if cb:
            callbacks.append(cb)
    args["stream"] = bool(callbacks)
    return args, {"callbacks": callbacks}


def make_llm_functions(config: Config) -> tuple[LLMFunctionType, LLMAsyncFunctionType]:
    _configure_open_ai_package(config)

    async def allm(prompt, **kwargs):
        args, options = _prepare_llm_arguments(config, kwargs)
        if is_chat_model(args["model"]):
            response = await openai.ChatCompletion.acreate(
                messages=prepare_chat_messages(prompt), **args
            )
            if args["stream"]:
                return await _a_process_streamed_response(
                    response, options["callbacks"], chat_model_used=True
                )

            for cb in options["callbacks"]:
                cb(response.choices[0].message.content)
            return LLMResponse(response.choices[0].message.content, response)

        response = await openai.Completion.acreate(
            prompt=prepare_prompt(prompt), **args
        )
        if args["stream"]:
            return await _a_process_streamed_response(
                response, options["callbacks"], chat_model_used=False
            )

        return LLMResponse(response.choices[0].text, response)

    def llm(prompt, **kwargs):
        args, options = _prepare_llm_arguments(config, kwargs)
        if is_chat_model(args["model"]):
            response = openai.ChatCompletion.create(
                messages=prepare_chat_messages(prompt), **args
            )
            if args["stream"]:
                return _process_streamed_response(
                    response, options["callbacks"], chat_model_used=True
                )

            for cb in options["callbacks"]:
                cb(response.choices[0].message.content)
            return LLMResponse(response.choices[0].message.content, response)

        # Else (if it is text completion model)
        response = openai.Completion.create(prompt=prepare_prompt(prompt), **args)
        if args["stream"]:
            return _process_streamed_response(
                response, options["callbacks"], chat_model_used=False
            )

        return LLMResponse(response.choices[0].text, response)

    return llm, allm



"""
# Minimalistic Foundation for AI Applications

**microcore** is a collection of python adapters for Large Language Models
and Semantic Search APIs allowing to
communicate with these services convenient way, make it easily switchable
and separate business logic from implementation details.
"""
import os
import microcore.ui  # noqa
from .embedding_db import SearchResult, AbstractEmbeddingDB
from .file_storage import storage
from ._env import configure, env, config
from .logging import use_logging
from .message_types import UserMsg, AssistantMsg, SysMsg, Msg
from .configuration import ApiType, LLMConfigError
from .types import BadAIJsonAnswer, BadAIAnswer
from .wrappers.prompt_wrapper import PromptWrapper
from .wrappers.llm_response_wrapper import LLMResponse
from ._llm_functions import llm, allm
from .utils import parse


def tpl(file: os.PathLike[str] | str, **kwargs) -> str | PromptWrapper:
    """Renders a prompt template using the provided parameters."""
    return PromptWrapper(env().tpl_function(file, **kwargs))


def prompt(template_str: str, **kwargs) -> str | PromptWrapper:
    """Renders a prompt template from string using the provided parameters."""
    return PromptWrapper(env().jinja_env.from_string(template_str).render(**kwargs))


def use_model(name: str):
    """Switches language model"""
    config().MODEL = name
    config().LLM_DEFAULT_ARGS["model"] = name


def validate_config():
    """
    Validates current MicroCore configuration

    Raises:
        `LLMConfigError` if configuration is invalid
    """
    config().validate()


class _EmbeddingDBProxy(AbstractEmbeddingDB):
    def get_all(self, collection: str) -> list[str | SearchResult]:
        return env().texts.get_all(collection)

    def search(
        self,
        collection: str,
        query: str | list,
        n_results: int = 5,
        where: dict = None,
        **kwargs,
    ) -> list[str | SearchResult]:
        return env().texts.search(collection, query, n_results, where, **kwargs)

    def save_many(self, collection: str, items: list[tuple[str, dict] | str]):
        return env().texts.save_many(collection, items)

    def save(self, collection: str, text: str, metadata: dict = None):
        return env().texts.save(collection, text, metadata)

    def clear(self, collection: str):
        return env().texts.clear(collection)


texts = _EmbeddingDBProxy()
"""Embedding database, see `microcore.embedding_db.AbstractEmbeddingDB`"""

__all__ = [
    "llm",
    "allm",
    "tpl",
    "texts",
    "configure",
    "validate_config",
    "storage",
    "use_model",
    "use_logging",
    "env",
    "config",
    "Msg",
    "UserMsg",
    "SysMsg",
    "AssistantMsg",
    "ApiType",
    "BadAIJsonAnswer",
    "BadAIAnswer",
    "LLMConfigError",
    "LLMResponse",
    "PromptWrapper",
    "parse",
    # submodules
    "embedding_db",
    "file_storage",
    "message_types",
    "utils",
    "configuration",
    "types",
    "ui",
    # "wrappers",
]

__version__ = "0.7.15"



import json
from typing import Any

from ..types import BadAIJsonAnswer
from ..utils import ExtendedString
from ..message_types import Role


def remove_json_wrapper(input_string: str) -> str:
    input_string = str(input_string).strip()
    if input_string.startswith("```json") and input_string.endswith("```"):
        json_content = input_string[7:-3].strip()
        return json_content

    return input_string


class LLMResponse(ExtendedString):
    """
    Response from the Large Language Model.

    If treated as a string, it returns the text generated by the LLM.

    Also, it contains all fields returned by the API accessible as an attributes.

    See fields returned by the OpenAI:

    - https://platform.openai.com/docs/api-reference/completions/object
    - https://platform.openai.com/docs/api-reference/chat/object
    """

    def __new__(cls, string: str, attrs: dict = None):
        obj = ExtendedString.__new__(cls, string, attrs)
        # Same fields like in Msg
        setattr(obj, "role", Role.ASSISTANT)
        setattr(obj, "content", str(string))
        return obj

    def parse_json(
        self, raise_errors: bool = True, required_fields: list[str] = None
    ) -> list | dict | Any:
        try:
            res = json.loads(remove_json_wrapper(self.content))
            if required_fields:
                if not isinstance(res, dict):
                    raise BadAIJsonAnswer("Not an object")
                for field in required_fields:
                    if field not in res:
                        raise BadAIJsonAnswer(f'Missing field "{field}"')
            return res
        except json.decoder.JSONDecodeError as e:
            if raise_errors:
                raise BadAIJsonAnswer() from e
            return False



from typing import Callable, Any, Awaitable, Union
from os import PathLike

TplFunctionType = Callable[[Union[PathLike[str], str], Any], str]
"""Function type for rendering prompt templates"""
LLMFunctionType = Callable[[str, Any], str]
"""Function type for requesting LLM synchronously"""
LLMAsyncFunctionType = Callable[[str, Any], Awaitable[str]]
"""Function type for requesting LLM asynchronously"""


class BadAIAnswer(ValueError):
    """Unprocessable response generated by the LLM"""

    def __init__(self, message: str = None, details: str = None):
        self.message = message or "Unprocessable response generated by the LLM"
        self.details = details
        super().__init__(message)

    def __str__(self):
        return self.message + f": {self.details}" if self.details else ""


class BadAIJsonAnswer(BadAIAnswer):
    def __init__(
        self, message: str = "Invalid JSON generated by the LLM", details=None
    ):
        super().__init__(message, details)



import builtins
import dataclasses
import inspect
import json
import os
import re
from fnmatch import fnmatch
from pathlib import Path

from .types import BadAIAnswer


def is_chat_model(model: str) -> bool:
    """Detects if model is chat model or text completion model"""
    completion_keywords = ["instruct", "davinci", "babbage", "curie", "ada"]
    return not any(keyword in model.lower() for keyword in completion_keywords)


class ExtendedString(str):
    """
    Provides a way of extending string with attributes and methods
    """

    def __new__(cls, string: str, attrs: dict = None):
        """
        Allows string to have attributes.
        """
        obj = str.__new__(cls, string)
        if attrs:
            for k, v in attrs.items():
                setattr(obj, k, v)
        return obj

    def __getattr__(self, item):
        """
        Provides chaining of global functions
        """
        global_func = inspect.currentframe().f_back.f_globals.get(item) or vars(
            builtins
        ).get(item, None)
        if callable(global_func):

            def method_handler(*args, **kwargs):
                res = global_func(self, *args, **kwargs)
                if isinstance(res, str) and not isinstance(res, ExtendedString):
                    res = ExtendedString(res)
                return res

            return method_handler

        # If there's not a global function with that name, raise an AttributeError as usual
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{item}'"
        )


class DataclassEncoder(json.JSONEncoder):
    """@private"""

    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)


json.JSONEncoder.default = DataclassEncoder().default


def parse(
    text: str, field_format: str = r"\[\[(.*?)\]\]", required_fields: list = None
) -> dict:
    """
    Parse a document divided into sections and convert it into a dictionary.
    """
    pattern = rf"{field_format}\n(.*?)(?=\n{field_format}|$)"
    matches = re.findall(pattern, text, re.DOTALL)
    result = {key.strip().lower(): value for key, value, _ in matches}
    if required_fields:
        for field in required_fields:
            if field not in result:
                raise BadAIAnswer(f"Field '{field}' is required but not found")
    return result


def file_link(file_path: str | Path):
    """Returns file name in format displayed in PyCharm console as a link."""
    return "file:///" + str(Path(file_path).absolute()).replace("\\", "/")


def list_files(
    target_dir: str | Path = "",
    exclude: list[str | Path] = None,
    relative_to: str | Path = None,
    absolute: bool = False,
    posix: bool = False,
) -> list[Path]:
    """
    Lists files in a specified directory, excluding those that match given patterns.

    This function traverses the specified directory recursively and returns a list of all files
    that do not match the specified exclusion patterns. It can return absolute paths,
    paths relative to the target directory, or paths relative to a specified directory.

    Args:
        target_dir (str | Path): The directory to search in.
        exclude (list[str | Path]): Patterns of files to exclude.
        relative_to (str | Path, optional): Base directory for relative paths.
            If None, paths are relative to `target_dir`. Defaults to None.
        absolute (bool, optional): If True, returns absolute paths. Defaults to False.
        posix (bool, optional): If True, returns posix paths. Defaults to False.

    Returns:
        list[Path]: A list of Path objects representing the files found.

    Example:
        exclude_patterns = ['*.pyc', '__pycache__/*']
        target_directory = '/path/to/target'
        files = list_files(target_directory, exclude_patterns)
    """
    exclude = exclude or []
    target = Path(target_dir or os.getcwd()).resolve()
    relative_to = Path(relative_to).resolve() if relative_to else None
    if absolute and relative_to is not None:
        raise ValueError(
            "list_files(): Cannot specify both 'absolute' and 'relative_to'. Choose one."
        )
    return [
        p.as_posix() if posix else p
        for p in (
            path.resolve() if absolute else path.relative_to(relative_to or target)
            for path in target.rglob("*")
            if path.is_file()
            and not any(
                fnmatch(str(path.relative_to(target)), str(pattern))
                for pattern in exclude
            )
        )
    ]



import os
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import Any

import dotenv

_MISSING = object()

TRUE_VALUES = ["1", "TRUE", "YES", "ON", "ENABLED"]
"""@private"""


def from_env(default=None):
    """
    Provides default value for the configuration dataclass
    from the environment variable with the name equal to field name in upper case"""
    return field(default=_MISSING, metadata=dict(_from_env=True, _default=default))


def get_bool_from_env(env_var: str, default: bool = False) -> bool:
    """Convert value of environment variable to boolean"""
    return os.getenv(env_var, str(default)).upper() in TRUE_VALUES


class ApiType:
    """LLM API types"""

    OPEN_AI = "open_ai"
    AZURE = "azure"
    """See https://learn.microsoft.com/en-us/azure/ai-services/openai/concepts/models"""
    LLM = "llm"
    ANYSCALE = "anyscale"
    """See https://www.anyscale.com/endpoints"""
    DEEP_INFRA = "deep_infra"
    """List of text generation models: https://deepinfra.com/models?type=text-generation"""


_default_dotenv_loaded = False


@dataclass
class BaseConfig:
    """Base class for configuration dataclasses"""

    USE_DOT_ENV: bool = None
    DOT_ENV_FILE: str | Path = None

    def __post_init__(self):
        self._dot_env_setup()
        self._update_from_env()

    def _dot_env_setup(self):
        global _default_dotenv_loaded

        if self.USE_DOT_ENV is None:
            self.USE_DOT_ENV = get_bool_from_env("USE_DOT_ENV", True)

        if self.USE_DOT_ENV:
            if self.DOT_ENV_FILE or not _default_dotenv_loaded:
                dotenv.load_dotenv(override=True, dotenv_path=self.DOT_ENV_FILE)
            if not self.DOT_ENV_FILE:
                _default_dotenv_loaded = True

    def _update_from_env(self):
        for f in fields(self):
            if f.metadata.get("_from_env") and getattr(self, f.name) is _MISSING:
                setattr(self, f.name, os.getenv(f.name.upper(), f.metadata["_default"]))


@dataclass
class _OpenAIEnvVars:
    # OS Environment variables expected by OpenAI library
    # Will be used as defaults for LLM
    # @todo: implement lib_defaults to take default values from openai lib if available
    OPENAI_API_TYPE: str = from_env(ApiType.OPEN_AI)
    OPENAI_API_KEY: str = from_env()
    OPENAI_API_BASE: str = from_env("https://api.openai.com/v1")
    OPENAI_API_VERSION: str = from_env()


@dataclass
class LLMConfig(BaseConfig, _OpenAIEnvVars):
    """LLM configuration"""

    LLM_API_TYPE: str = from_env()
    """
    See `ApiType`.
    To use services that is not listed in `ApiType`,
    but provides OpenAPI interface, use `ApiType.OPEN_AI`"""

    LLM_API_KEY: str = from_env()
    LLM_API_BASE: str = from_env()
    """Base URL for the LLM API, e.g. https://api.openai.com/v1"""

    LLM_API_VERSION: str = from_env()
    LLM_DEPLOYMENT_ID: str = from_env()
    """Required by `ApiType.AZURE`"""

    MODEL: str = from_env()
    """Language model name"""

    LLM_DEFAULT_ARGS: dict = field(default_factory=dict)
    """
    You may specify here default arguments for the LLM API calls,
     i. e. temperature, max_tokens, etc.
     """

    AZURE_DEPLOYMENT_ID: str = from_env()

    def __post_init__(self):
        super().__post_init__()
        self._init_llm_options()
        self.validate()

    def _init_llm_options(self):
        # Use defaults from ENV variables expected by OpenAI API
        self.LLM_API_TYPE = self.LLM_API_TYPE or self.OPENAI_API_TYPE
        self.LLM_API_KEY = self.LLM_API_KEY or self.OPENAI_API_KEY
        self.LLM_API_BASE = self.LLM_API_BASE or self.OPENAI_API_BASE
        self.LLM_API_VERSION = self.LLM_API_VERSION or self.OPENAI_API_VERSION

        if self.LLM_API_TYPE == ApiType.AZURE:
            self.LLM_DEPLOYMENT_ID = self.LLM_DEPLOYMENT_ID or self.AZURE_DEPLOYMENT_ID

        if self.LLM_API_TYPE == ApiType.ANYSCALE:
            self.LLM_API_BASE = (
                self.LLM_API_BASE or "https://api.endpoints.anyscale.com/v1"
            )
            self.MODEL = self.MODEL or "meta-llama/Llama-2-70b-chat-hf"

        if self.LLM_API_TYPE == ApiType.DEEP_INFRA:
            self.LLM_API_BASE = (
                self.LLM_API_BASE or "https://api.deepinfra.com/v1/openai"
            )
            self.MODEL = self.MODEL or "meta-llama/Llama-2-70b-chat-hf"

        self.MODEL = self.MODEL or "gpt-3.5-turbo"

    def validate(self):
        """
        Validate LLM configuration

        Raises:
            LLMConfigError
        """
        if not self.LLM_API_KEY:
            raise LLMConfigError("LLM configuration error: LLM_API_KEY is absent")
        if self.LLM_API_TYPE == ApiType.AZURE:
            if not self.LLM_API_BASE:
                raise LLMConfigError(
                    "LLM configuration error: "
                    "LLM_API_BASE is required for using Azure models"
                )
            if not self.LLM_DEPLOYMENT_ID:
                raise LLMConfigError(
                    "LLM configuration error: "
                    "LLM_DEPLOYMENT_ID is required for using Azure models"
                )
            if not self.LLM_API_VERSION:
                raise LLMConfigError(
                    "LLM configuration error: "
                    "LLM_API_VERSION is required for using Azure models"
                )


class LLMConfigError(ValueError):
    """LLM configuration error"""


@dataclass
class Config(LLMConfig):
    """MicroCore configuration"""

    USE_LOGGING: bool = False
    """Whether to use logging or not, see `microcore.use_logging`"""

    PROMPT_TEMPLATES_PATH: str | Path = from_env("tpl")
    """Path to the folder with prompt templates, ./tpl by default"""

    STORAGE_PATH: str | Path = from_env("storage")
    """Path to the folder with file storage, ./storage by default"""

    STORAGE_DEFAULT_FILE_EXT: str = field(default="")

    EMBEDDING_DB_FOLDER: str = "embedding_db"
    """Folder within microcore.config.Config.STORAGE_PATH for storing embeddings"""

    EMBEDDING_DB_FUNCTION: Any = from_env()

    EMBEDDING_DB_ALLOW_DUPLICATES: bool = False

    DEFAULT_ENCODING: str = from_env("utf-8")
    """Used in file system operations, utf-8 by default"""

    JINJA2_AUTO_ESCAPE: bool = None

    def __post_init__(self):
        if self.JINJA2_AUTO_ESCAPE is None:
            self.JINJA2_AUTO_ESCAPE = get_bool_from_env("JINJA2_AUTO_ESCAPE", False)
        super().__post_init__()



from .message_types import DEFAULT_MESSAGE_ROLE


def prepare_prompt(prompt) -> str:
    """Converts prompt to string for LLM completion API"""
    return "\n".join(
        [
            str(p["content"]) if isinstance(p, dict) and "content" in p else str(p)
            for p in (prompt if isinstance(prompt, list) else [prompt])
        ]
    )


def prepare_chat_messages(prompt):
    """Converts prompt to messages for LLM chat API (OpenAI)"""
    return [
        dict(role=DEFAULT_MESSAGE_ROLE, content=msg) if isinstance(msg, str) else msg
        for msg in (prompt if isinstance(prompt, list) else [prompt])
    ]



from dataclasses import dataclass, field
from importlib.util import find_spec
import jinja2

from .configuration import Config
from . import AbstractEmbeddingDB
from .types import TplFunctionType, LLMAsyncFunctionType, LLMFunctionType
from .templating.jinja2 import make_jinja2_env, make_tpl_function
from .llm.openai_llm import make_llm_functions


@dataclass
class Env:
    config: Config = field(default_factory=Config)
    jinja_env: jinja2.Environment = None
    tpl_function: TplFunctionType = None
    llm_async_function: LLMAsyncFunctionType = None
    llm_function: LLMFunctionType = None
    llm_before_handlers: list[callable] = field(default_factory=list)
    llm_after_handlers: list[callable] = field(default_factory=list)
    texts: AbstractEmbeddingDB = None

    def __post_init__(self):
        global _env
        _env = self
        self.init_templating()
        self.init_llm()
        if self.config.USE_LOGGING:
            from .logging import use_logging

            use_logging()
        self.init_similarity_search()

    def init_templating(self):
        self.jinja_env = make_jinja2_env(self)
        self.tpl_function = make_tpl_function(self)

    def init_llm(self):
        self.llm_function, self.llm_async_function = make_llm_functions(self.config)

    def init_similarity_search(self):
        if find_spec("chromadb") is not None:
            from .embedding_db.chromadb import ChromaEmbeddingDB

            self.texts = ChromaEmbeddingDB(self.config)


@dataclass
class _Configure(Config):
    def __post_init__(self):
        super().__post_init__()
        Env(self)


configure: callable = _Configure
"""Applies configuration to MicroCore environment"""

_env: Env | None = None


def env() -> Env:
    """Returns the current MicroCore environment"""
    return _env or Env()


def config() -> Config:
    """Resolve current configuration"""
    return env().config



import builtins
import yaml
from jinja2 import PackageLoader
from ._env import env

_original_import = builtins.__import__


def custom_import(name, global_vars=None, local_vars=None, fromlist=(), level=0):
    module = _original_import(name, global_vars, local_vars, fromlist, level)
    if module.__doc__ and "ai_module" in module.__doc__:
        data = yaml.safe_load(module.__doc__)
        tpl_path = data.get("tpl_path", "")
        env().jinja_env.loader.loaders.append(PackageLoader(module.__name__, tpl_path))
    return module


builtins.__import__ = custom_import



from .message_types import Msg
from .wrappers.llm_response_wrapper import LLMResponse
from ._env import env


def llm(prompt: str | Msg | list[str] | list[Msg], **kwargs) -> str | LLMResponse:
    """
    Request Large Language Model synchronously

    Args:
        prompt (str | list[str]): Text to send to LLM
        **kwargs (dict): Parameters supported by the LLM API

            See parameters supported by the OpenAI:

            - https://platform.openai.com/docs/api-reference/completions/create
            - https://platform.openai.com/docs/api-reference/chat/create

            **Additional parameters:**

                - callback: callable - callback function
                to be called on each chunk of text,
                enables response streaming if supported by the LLM API
                - callbacks: list[callable] - collection of callbacks
                to be called on each chunk of text,
                enables response streaming if supported by the LLM API

    Returns:

        Text generated by the LLM as string
        with all fields returned by API accessible as an attributes.

        See fields returned by the OpenAI:

        - https://platform.openai.com/docs/api-reference/completions/object
        - https://platform.openai.com/docs/api-reference/chat/object
    """
    [h(prompt, **kwargs) for h in env().llm_before_handlers]
    response = env().llm_function(prompt, **kwargs)
    [h(response) for h in env().llm_after_handlers]
    return response


async def allm(
    prompt: str | Msg | list[str] | list[Msg], **kwargs
) -> str | LLMResponse:
    """
    Request Large Language Model asynchronously

    Args:
        prompt (str | list[str]): Text to send to LLM
        **kwargs (dict): Parameters supported by the LLM API

            See parameters supported by the OpenAI:

            - https://platform.openai.com/docs/api-reference/completions/create
            - https://platform.openai.com/docs/api-reference/chat/create

            **Additional parameters:**

            - callback: callable - callback function
            to be called on each chunk of text,
            enables response streaming if supported by the LLM API
            - callbacks: list[callable] - collection of callbacks
            to be called on each chunk of text,
            enables response streaming if supported by the LLM API

            Note: async callbacks are supported only for async LLM API calls

    Returns:

        Text generated by the LLM as string
        with all fields returned by API accessible as an attributes.

        See fields returned by the OpenAI:

        - https://platform.openai.com/docs/api-reference/completions/object
        - https://platform.openai.com/docs/api-reference/chat/object
    """
    [h(prompt, **kwargs) for h in env().llm_before_handlers]
    response = await env().llm_async_function(prompt, **kwargs)
    [h(response) for h in env().llm_after_handlers]
    return response



import dataclasses
from colorama import Fore, Style, init

from .configuration import ApiType
from ._env import env, config
from ._prepare_llm_args import prepare_chat_messages, prepare_prompt
from .utils import is_chat_model


class LoggingConfig:
    PROMPT_COLOR = Fore.LIGHTGREEN_EX
    RESPONSE_COLOR = Fore.CYAN
    INDENT: str = "\t"
    DENSE: bool = False


def _log_request(prompt, **kwargs):
    nl = "\n" if LoggingConfig.DENSE else "\n" + LoggingConfig.INDENT
    model = _resolve_model(**kwargs)
    print(
        f"Requesting LLM {Fore.MAGENTA}{model}{Style.RESET_ALL}:",
        end=" " if LoggingConfig.DENSE else "\n",
    )
    if is_chat_model(model):
        for msg in prepare_chat_messages(prompt):
            role, content = (
                (msg["role"], msg["content"])
                if isinstance(msg, dict)
                else dataclasses.astuple(msg)
            )
            nl2 = "\n" if LoggingConfig.DENSE else nl + LoggingConfig.INDENT
            content = (" " if LoggingConfig.DENSE else nl2) + nl2.join(
                content.split("\n")
            )
            print(
                f'{"" if LoggingConfig.DENSE else LoggingConfig.INDENT}'
                f"{LoggingConfig.PROMPT_COLOR}[{role.capitalize()}]:{content}"
            )
    else:
        lines = prepare_prompt(prompt).split("\n")
        print(
            LoggingConfig.PROMPT_COLOR
            + (" " if LoggingConfig.DENSE else LoggingConfig.INDENT)
            + nl.join(lines)
        )


def _resolve_model(**kwargs):
    cfg = config()
    model = kwargs.get("model") or cfg.LLM_DEFAULT_ARGS.get("model") or cfg.MODEL
    if cfg.LLM_API_TYPE == ApiType.AZURE:
        model = f"azure:{model}"
    return model


def _log_response(out):
    nl = "\n" if LoggingConfig.DENSE else "\n" + LoggingConfig.INDENT
    out_indented = (" " if LoggingConfig.DENSE else nl) + nl.join(out.split("\n"))
    print(f"LLM Response:{LoggingConfig.RESPONSE_COLOR}{out_indented}")


def use_logging():
    """Turns on logging of LLM requests and responses to console."""
    init(autoreset=True)
    if _log_request not in env().llm_before_handlers:
        env().llm_before_handlers.append(_log_request)
    if _log_response not in env().llm_after_handlers:
        env().llm_after_handlers.append(_log_response)



from dataclasses import dataclass

import chromadb
import uuid
from chromadb.config import Settings
from chromadb.utils import embedding_functions
from ..configuration import Config
from .. import SearchResult, AbstractEmbeddingDB


@dataclass
class ChromaEmbeddingDB(AbstractEmbeddingDB):
    config: Config
    embedding_function: embedding_functions.EmbeddingFunction = None
    client: chromadb.Client = None

    def __post_init__(self):
        self.client = chromadb.PersistentClient(
            path=f"{self.config.STORAGE_PATH}/{self.config.EMBEDDING_DB_FOLDER}",
            settings=Settings(anonymized_telemetry=False),
        )
        self.embedding_function = (
            self.config.EMBEDDING_DB_FUNCTION
            or embedding_functions.DefaultEmbeddingFunction()
        )

    @classmethod
    def _wrap_results(cls, results) -> list[str | SearchResult]:
        return [
            SearchResult(
                results["documents"][0][i],
                dict(
                    metadata=results["metadatas"][0][i] or {},
                    id=results["ids"][0][i],
                    distance=results["distances"][0][i],
                ),
            )
            for i in range(len(results["documents"][0]))
        ]

    def search(
        self,
        collection: str,
        query: str | list,
        n_results: int = 5,
        where: dict = None,
        **kwargs,
    ) -> list[str | SearchResult]:
        try:
            chroma_collection = self.client.get_collection(
                collection, embedding_function=self.embedding_function
            )
        except ValueError:
            return []

        if isinstance(query, str):
            query = [query]

        d = chroma_collection.query(
            query_texts=query, n_results=n_results, where=where, **kwargs
        )
        return (
            self._wrap_results(d)
            if d and d.get("documents") and d["documents"][0]
            else []
        )

    def save_many(self, collection: str, items: list[tuple[str, dict] | str]):
        chroma_collection = self.client.get_or_create_collection(
            name=collection, embedding_function=self.embedding_function
        )
        unique = not self.config.EMBEDDING_DB_ALLOW_DUPLICATES
        texts, ids, metadatas = [], [], []
        for i in items:
            if isinstance(i, str):
                text = i
                metadata = None
            else:
                text = i[0]
                metadata = i[1] or None
            if unique and text in texts:
                continue
            texts.append(text)
            metadatas.append(metadata)
            ids.append(str(hash(text)) if unique else str(uuid.uuid4()))
        chroma_collection.upsert(documents=texts, ids=ids, metadatas=metadatas)

    def clear(self, collection: str):
        try:
            self.client.delete_collection(collection)
        except ValueError:
            pass

    def get_all(self, collection: str) -> list[str | SearchResult]:
        try:
            chroma_collection = self.client.get_collection(
                collection, embedding_function=self.embedding_function
            )
        except ValueError:
            return []
        results = chroma_collection.get()
        return [
            SearchResult(
                results["documents"][i],
                {"metadata": results["metadatas"][i] or {}, "id": results["ids"][i]},
            )
            for i in range(len(results["documents"]))
        ]
